require(Matrix)
require(abind)
#' Simulates the COVID-19 dynamics of interacting populations
#' @param R0default The default R0 value wih no mititgations
#' @param R0init The initial R0 value for the unrestricted population. Useful for modeling homogeneous mitigations. 
#' @param popParams A data.frame of population bracket parameters.  Must contain "age" (character name for the age bracket),"hosp" (fraction of cases needing hospitalization), "crit" (fraction of hospitalizations needing ICU), "mort" case mortality, "fraction" (fraction of the population in that bracket). Example in populationParams.txt. 
#' @param restriction.effect=.3 The mulitplier for interaction frequency applied to the contact matrix when mitigations are present
#' @param free.bracket The bracket ID for the unrestricted population. All brackets up to free.bracket are unrestricted at time 0.
#' @param free.bracket.fraction=1 the fraction of the population in brackets up.to free.bracket that is unrestricted. Some inidividuals are still restricted because of risk factors other than age.
#'  @param interactionMat The unitless, symmetric interaction frequency matrix for population brackets defined in popParams. Example in interactionMatrix.txt.
#'  @param step.up.days Length in days of the ramp up period to normal transmision
#'  @param forceNormalAt Time in days when normal transmission resumes
#'  @param  step.size Step size in days
#'  @param nstep.max Maximum number of steps
#'  @param nstep.min Minimum number of steps
#' @examples




simulateSIRmultiPopICU=function(R0default,  R0init, popParams, restriction.effect=.3, free.bracket=4, free.bracket.fraction=1, interactionMat=NULL,  step.up.days=90, forceNormalAt=NULL,  step.size=1, nstep.max=365*3, nstep.min=NULL){
  
  if(is.null(interactionMat)){ #default interaction matrix is just all 1s, all interactions are equally likely
    interactionMat=matrix(1, nrow=nrow(popParams), ncol=nrow(popParams)) 
  }
  #number of populations we are modeling
  npop=nrow(popParams)
  
  #hard coded parameters"
  I=100000 #total number of infected individuals
  year=365 #year
  recoverytime=14 #recovery time in days. Taken to be 14 day
  N=totalPop=308745538 #US population, last census
  
  
  iif=1:free.bracket #index of the free age bracket
  iir=(free.bracket+1):npop #index of the restricted bracket
  
  #if free.bracket fraction is less than 1 we have to split those age groups in two
  if(free.bracket.fraction<1){ 
    #set up expanded populations
    popParamsFreeR=popParamsFree=popParams[iif,]
    popParamsFreeR$age=paste(popParamsFreeR$age, "R")
    tmpfrac=popParamsFree$fraction
    #redistribute the population fraction  across free and free-R (technically free but restricted due to factors other than age)
    popParamsFree$fraction=tmpfrac*free.bracket.fraction 
    popParamsFreeR$fraction=tmpfrac*(1-free.bracket.fraction) 
    #this is the restricted brackets
    popParamsR=popParams[iir,]
    #the final popParams is a combination of free, free-R, and R
    popParamsThis=rbind(popParamsFree, popParamsFreeR,popParamsR)
    
    #expand the interaction matrix in the same way
    #the proportions are facored out of the interaction matrix so there is need to do any adjustments, we can just copy
    A=interactionMat[iif,iif]
    B=interactionMat[iif, iir]
    C=interactionMat[iir, iir]
    
    interactionMat=rbind(cbind(A,A,B), cbind(A,A,B), cbind(t(B),t(B),C))
    rownames(interactionMat)=colnames(interactionMat)=popParamsThis$age
    
  }
  else{ #otherwise popParams and interactionMat can be left as is
    popParamsThis=popParams
  }
  
  #adjust the number of population brackets
  npop=nrow(popParamsThis)
  
  #total number of individuals in each bracket
  N0Vec=N*popParamsThis$fraction
  
  
  
  
  
  #default values for gamma and delta variables that specify the rates of leaving the infected state
  alpha=1/recoverytime #total rate of leaving the infected state
  #alpha is split into two componenets
  #delta is the mortality rate
  dVec=popParamsThis$mort/14 #mortality per age group
  #gamma is the rate of recovery
  gVec=alpha-dVec
  
  
  #compute the default beta corresponding to R=Rdefault
  Qmat=interactionMat%*%diag(popParamsThis$fraction) # age distribution adjusted interaction matrix
  #the initial "steady state" distribution and the R0 multiplier are eigenvectors and eigenvalue of Q matrix
  ee=eigen(t(Qmat))
  ss=Re(ee$vectors[,1])
  ss=ss*sign(ss[1]) #make positive
  ss=ss/sum(ss) #normalize to sum=1
  beta0=R0default*alpha/(Re(ee$values[1]))
  show(beta0)
  
  #in order to adjust from R0 default we compute multiplicative adjustments
  #this is for global mitigations in the initial phase
  multInit=R0init/R0default
  
  
  
  #distribute infectoins according to steady state
  Ivec=I0Vec=I*ss
  #setting up initial conditions
  Svec=N0Vec-I0Vec #number susceptible
  Rvec=double(length(Svec))
  Mvec=double(length(Svec))
  
  
  #adjust the interaction matrix to account for mitigations
  interactionMatFree=interactionMat #interactionMatFree*beta0 will give R0default transmission
  interactionMat=interactionMatFree*restriction.effect*multInit #initial transmission
  interactionMat[iif, iif]=interactionMatFree[iif,iif]*multInit #release the free bracked [indexed by iff]
  interactionMatR=interactionMat #save the restricted interaction matrix for output, the interactionMat itself will change
  
  
  #so the output will be in order
  # S, R, I, M, H, C for each age group
  
  output=array(data = 0,dim=c(1, 6, nrow(popParamsThis)), dimnames = list("1", c("S", "R", "I", "M", "H", "C"), popParamsThis$age))
  
  #initialize infected and susceptible, everything else is 0
  output[1,"S",]=Svec
  output[1,"I",]=Ivec
  
  output[1, "H",]=Ivec*popParamsThis$hosp
  output[1, "C",]=Ivec*popParamsThis$crit*popParamsThis$hosp
  
  peak.reached=F
  peak.val=-1
  restrictions=T
  s=0
  rlifted=Inf
  smin=0
  
  if(is.null(nstep.min)){
    smin=forceNormalAt*1.25
  }
  else{
    smin=nstep.min
  }
  #stopping criteria is number of infections < 10000 or nstep.max is reached
  while((sum(output[dim(output)[1],"I",])>10000 && s<nstep.max)||s<smin){
    s=s+1 
    day=s/step.size
    
    
    
    
    curTime=output[dim(output)[1],,] #get the current slice
    
    #compute the next step according to the SIR model
    
    
    #  newI=curTime["I",]+((tmpIplus<-rowSums(outer(curTime["I",], curTime["S",])*beta0/N))-alpha*curTime["I",])*step.size
    
    newI=curTime["I",]+((tmpIplus<-(curTime["I",]%*%(interactionMat)%*%diag(curTime["S",])*beta0/N))-alpha*curTime["I",])*step.size
    
    #handle the forced normal point
    
    #begin lifting restrictions
    if(abs(forceNormalAt-day-step.up.days)<1e-6){
      rlifted=s
      message(paste("Starting at", s))
    }
    #ramp up perriod
    if(forceNormalAt-day<step.up.days && restrictions==T){
      #message(paste("Force normal step up", day))
      daydiff=day-rlifted
      interactionMat=interactionMatFree*(multInit*restriction.effect*(1-daydiff/step.up.days)+(daydiff/step.up.days)) #ramp up from init*restriction to release
      
      interactionMat[iif,iif]=(interactionMatFree*(multInit*(1-daydiff/step.up.days)+(daydiff/step.up.days)))[iif,iif] #ramp up from init to 1
      
      #show(interactionMat[8,8])
      restrictions=T
    }
    
    if(abs(day-forceNormalAt)<1e-6){
      message("Force normal final")
      interactionMat=interactionMatFree
      interactionMat[iif,iif]=interactionMatFree[iif,iif] #set0
      restrictions=F
    }
    
    #update the other variables
  newS=curTime["S",]-tmpIplus*step.size
  newR=curTime["R",]+gVec*curTime["I",]*step.size
  newM=curTime["M",]+dVec*curTime["I",]*step.size
  #create the next time step in order
  #c("S", "R", "I", "M", "H", "C")
  # variables H and C are not part of the model, they are computed from newI
  nextTime=rbind(newS, newR, newI, newM, newI*popParamsThis$hosp, newI*popParamsThis$crit*popParamsThis$hosp)
  rownames(nextTime)=c("S", "R", "I", "M", "H", "C")
  output=abind(output, nextTime, along=1)    #append to output
}
#show(c(max(apply(output[, "C", ],1,sum)), totalICU))
#save the relaxation time
stopifnot(s>=smin)
return(list(data=output, rlifted=rlifted, interactionMatR=interactionMatR, interactionMatFree=interactionMatFree, free.bracket=free.bracket, free.bracket.fraction=free.bracket.fraction, restriction.effect=restriction.effect, R0default=R0default, R0init=R0init,  step.up.days=step.up.days))

}
